#include <config.h>
! generate a list of three-dimensional paths
!   (M.Hayakawa, May 12, 2011)
!   [ modularize the part of previous version of "path_generate_class.F90"
!     which generates path according to the generalized Brensenham algorithm ]

module bresenham_module
  use constants_module
!--- module "param_size" was defined in HMCLIB/chrolog_class.F
!  use param_size
  use lattice_class
!  use error_class
  implicit none
  private

  public :: disp
  public :: new
  public :: delete
  public :: run

  type disp
    ! Array of directions of step vectors
    ! Each element, dir(j), takes either one of the values {1,2,3}
    ! dir(j) = 1 represents a step
    ! with the vector (a,0,0) in unit of lattice spacing, and so on.
    ! Size of array is equal to the total number of steps
    integer, allocatable :: dir(:) 
    ! pointer to the next disp object
    type(disp), pointer :: next => NULL()
  end type disp

  interface new
    module procedure new_disp
  end interface

  interface free
    module procedure free_disp
  end interface

  interface delete
    module procedure delete_disp
  end interface

  interface Bresenham
    module procedure Bresenham_2d
    module procedure Bresenham_3d
  end interface

  interface permute
    module procedure permute_2
    module procedure permute_3
  end interface

  interface sflip
    ! get new displacement vectors by flipping (specified) directions
    module procedure sflip_all
    module procedure sflip_1
    module procedure sflip_2
  end interface

  interface run
    module procedure run_3
    module procedure run_2
    module procedure run_1
  end interface

contains
!---------- IMPLEMENTATION --------------

subroutine new_disp( this, nstep )
!
! allocate "dir"
!
! inout:
!   this : disp object
!
! inout:
!   nstep : number of steps = number of step vectors
!
  implicit none
  type(disp), intent(inout) :: this
  integer :: nstep

  if ( allocated(this%dir) ) deallocate(this%dir)
  allocate( this%dir(nstep) )

  return
end subroutine new_disp


subroutine delete_disp( this )
!
! deconstructor
!
! inout:
!   this : disp object
!
  implicit none
  type(disp), intent(inout) :: this
!--------------------------------------------------

  if ( allocated(this%dir) ) deallocate(this%dir)
  call free( this%next )

  return
end subroutine delete_disp


recursive subroutine free_disp( dsp )
!
! free v in all lists
!
! inout:
!   dsp : disp object
!
  implicit none
  type(disp), intent(inout), pointer :: dsp

  if ( associated(dsp) ) then
    call free_disp( dsp%next )
    deallocate( dsp%dir )
    deallocate( dsp )
!    write(*,*) "free !"  ! just for debug purpose
  endif

  return
end subroutine free_disp


subroutine Bresenham_2d( this, lmax, lmin )
!
! generate displacement vectors via 2-dim verions of Bresenham algorithm 
!
! inout:
!   this : disp object
!
! input :
!   lmax : maximum size of displacement (x-direction)
!   lmin : minimum size of displacement (y-direction)
!
  type(disp), intent(inout) :: this
  integer, intent(in) :: lmax, lmin
  integer, parameter :: dir_max = 1, dir_min = 2
!----------------------------------------------------------
  integer :: lmax2, lmin2
  integer :: h
  integer :: iter
  integer :: istep, nstep

  nstep = lmax + lmin
  if ( nstep /= SIZE(this%dir) ) then
    call new( this, nstep )
  endif

  lmax2 = 2 *lmax
  lmin2 = 2 *lmin
  h = lmin2 - lmax

  istep = 0
  do iter = 1, lmax
    istep = istep + 1
    this%dir(istep) = dir_max ! max-direction
    if ( h >= 0 ) then
      h = h - lmax2
      istep = istep + 1
      this%dir(istep) = dir_min ! min-direction
    endif
    h = h + lmin2
  enddo ! iter
  
  return
end subroutine Bresenham_2d


subroutine Bresenham_3d( this, lmax, lmid, lmin )
!
! generate displacement vectors via 3-dim verions of Bresenham algorithm 
!
! inout:
!   this : disp object
!
! input :
!   lmax : maximum size of displacement (x-direction)
!   lmid : middle size of displacement  (y-direction)
!   lmin : minimum size of displacement (z-direction)
!
  type(disp), intent(inout) :: this
  integer, intent(in) :: lmax, lmid, lmin
  integer, parameter :: dir_max = 1, dir_mid = 2, dir_min = 3
!----------------------------------------------------------
  integer :: lmax2, lmid2, lmin2
  integer :: h_max_mid, h_max_min
  integer :: iter
  integer :: istep, nstep

  nstep = lmax + lmid + lmin
  if ( nstep /= SIZE(this%dir) ) then
    call new( this, nstep )
  endif

  lmax2 = 2 *lmax
  lmid2 = 2 *lmid
  lmin2 = 2 *lmin
  h_max_mid = lmid2 - lmax
  h_max_min = lmin2 - lmax

  istep = 0
  do iter = 1, lmax
    istep = istep + 1
    this%dir(istep) = dir_max ! max-direction
    if ( h_max_mid >= 0 ) then
      h_max_mid = h_max_mid - lmax2
      istep = istep + 1
      this%dir(istep) = dir_mid ! mid-direction
    endif
    if ( h_max_min >= 0 ) then
      h_max_min = h_max_min - lmax2
      istep = istep + 1
      this%dir(istep) = dir_min ! min-direction
    endif
    h_max_mid = h_max_mid + lmid2
    h_max_min = h_max_min + lmin2
  enddo ! iter
  
  return
end subroutine Bresenham_3d


subroutine permute_2( this, src, dir1, dir2 )
!
! generate new displacement vectors by dir1 <=> dir2 
!
! inout:
!   this : disp object
!
! input:
!   src  : disp object
!   dir1 : direction \in {1,2,3}
!   dir2 : direction (\ne dir1) \in {1,2,3}
!
  type(disp), intent(inout) :: this
  type(disp), intent(in) :: src
  integer, intent(in) :: dir1, dir2
!------------------------------------------------------
  integer :: j

  do j = 1, SIZE(this%dir)
    if ( src%dir(j) == dir1 ) then
      this%dir(j) = dir2
    else if ( src%dir(j) == dir2 ) then
      this%dir(j) = dir1
    else
      this%dir(j) = src%dir(j)
    endif
  enddo

  return
end subroutine permute_2


subroutine permute_3( this, src, dir1, dir2, dir3 )
!
! generate new displacement vectors by 
!   dir1 <= (1 in src), dir2 <= (2 in src) dir3 <= (3 in src)
!
! inout:
!   this : disp object
!
! input:
!   src  : disp object
!   dir1 : direction \in {1,2,3}
!   dir2 : (\ne dir1) \in {1,2,3}
!   dir3 : (\ne dir1, dir2) \in {1,2,3}
!
  type(disp), intent(inout) :: this
  type(disp), intent(in) :: src
  integer, intent(in) :: dir1, dir2, dir3
!------------------------------------------------------
  integer :: j

  do j = 1, SIZE(this%dir)
    if ( src%dir(j) == 1 ) then
      this%dir(j) = dir1
    else if ( src%dir(j) == 2 ) then
      this%dir(j) = dir2
    else
      this%dir(j) = dir3
    endif
  enddo

  return
end subroutine permute_3

subroutine sflip_all( this, src )
!
! generate new displacement vectors by flipping all directions 
!
! inout:
!   this : disp object
!
! input:
!   src  : disp object
!
  type(disp), intent(inout) :: this
  type(disp), intent(in) :: src
!------------------------------------------------------
  integer :: j

  do j = 1, SIZE(this%dir)
    this%dir(j) = - src%dir(j)
  enddo

  return
end subroutine sflip_all


subroutine sflip_1( this, src, dir )
!
! generate new displacement vectors by flipping the sign of direction, dir 
!
! inout:
!   this : disp object
!
! input:
!   src : disp object
!   dir : direction to be reversed
!
  type(disp), intent(inout) :: this
  type(disp), intent(in) :: src
  integer, intent(in) :: dir
!------------------------------------------------------
  integer :: j

  do j = 1, SIZE(this%dir)
    if ( src%dir(j) == dir ) then
      this%dir(j) = - src%dir(j)
    else
     this%dir(j) = src%dir(j)
    endif
  enddo

  return
end subroutine sflip_1


subroutine sflip_2( this, src, dir1, dir2 )
!
! generate new displacement vectors by flipping the signs of directions,
! dir1, dir2 
!
! inout:
!   this : disp object
!
! input:
!   src  : disp object
!   dir1 : direction to be reversed
!   dir2 : direction to be reversed (\ne dir1)
!
  type(disp), intent(inout) :: this
  type(disp), intent(in) :: src
  integer, intent(in) :: dir1, dir2
!------------------------------------------------------
  integer :: j

  do j = 1, SIZE(this%dir)
    if ( ( src%dir(j) == dir1 ) .or. ( src%dir(j) == dir2 ) ) then
      this%dir(j) = - src%dir(j)
    else
      this%dir(j) = src%dir(j)
    endif
  enddo

  return
end subroutine sflip_2


subroutine run_3( this, lmax, lmid, lmin, last_p )
!
! generate all possible displacement vectors 
! upon the displacement vector obtained via 3-dim verions
! of Bresenham algorithm
!
! Caution : 
!  (1) only use for lmin \ne 0
!      (Otherwise, generate patterns include the same ones)
!  (2) user is responsible to correctly give lmax >= lmid >= lmin
!
! inout:
!   this : disp object
!
! input :
!   lmax : maximum size of displacement (x-direction)
!   lmid : middle size of displacement  (y-direction)
!   lmin : minimum size of displacement (z-direction)
!
! output:
!   last_p : pointer to the last disp object in the list (optional)
!
  type(disp), intent(inout), target :: this
  integer, intent(in) :: lmax, lmid, lmin
  type(disp), intent(out), optional, pointer :: last_p
!------------------------------------------------------
  type(disp), pointer :: dsp_p
  integer :: isize

  ! generate basic steps 
  call Bresenham( this, lmax, lmid, lmin )

  isize = SIZE(this%dir)
  dsp_p => this

  ! (1,3,2)
  allocate( dsp_p%next )
  dsp_p => dsp_p%next
  call new( dsp_p, isize )
  call permute( dsp_p, this, 2, 3 )

  ! (2,1,3)
  allocate( dsp_p%next )
  dsp_p => dsp_p%next
  call new( dsp_p, isize )
  call permute( dsp_p, this, 1, 2 )

  ! (2,3,1) 
  allocate( dsp_p%next )
  dsp_p => dsp_p%next
  call new( dsp_p, isize )
  call permute( dsp_p, this, 2, 3, 1 )

  ! (3,1,2)
  allocate( dsp_p%next )
  dsp_p => dsp_p%next
  call new( dsp_p, isize )
  call permute( dsp_p, this, 3, 1, 2 )

  ! (3,2,1)
  allocate( dsp_p%next )
  dsp_p => dsp_p%next
  call new( dsp_p, isize )
  call permute( dsp_p, this, 3, 1 )

  if ( present(last_p) ) then
    last_p => dsp_p
  endif

  return
end subroutine run_3


subroutine run_2( this, lmax, lmin, last_p )
!
! generate all possible displacement vectors 
! upon the displacement vector obtained via 3-dim verions
! of Bresenham algorithm
!
! Caution : 
!  (1) only use for lmin \ne 0
!      (Otherwise, generate patterns include the same ones)
!  (2) user is responsible to correctly give lmax >= lmin
!
! inout:
!   this : disp object
!
! input :
!   lmax : maximum size of displacement (x-direction)
!   lmin : middle size of displacement  (y-direction)
!
! output:
!   last_p : pointer to the last disp object in the list (optional)
!
  type(disp), intent(inout), target :: this
  integer, intent(in) :: lmax, lmin
  type(disp), intent(out), optional, pointer :: last_p
!------------------------------------------------------
  type(disp), pointer :: dsp_p
  integer :: isize

  ! generate basic steps
  call Bresenham( this, lmax, lmin )

  isize = SIZE(this%dir)
  dsp_p => this

  ! (1,3,2) (no 2)
  allocate( dsp_p%next )
  dsp_p => dsp_p%next
  call new( dsp_p, isize )
  call permute( dsp_p, this, 2, 3 )

  ! (2,1,3) (no 3)
  allocate( dsp_p%next )
  dsp_p => dsp_p%next
  call new( dsp_p, isize )
  call permute( dsp_p, this, 1, 2 )

  ! (2,3,1) (no 1)
  allocate( dsp_p%next )
  dsp_p => dsp_p%next
  call new( dsp_p, isize )
  call permute( dsp_p, this, 2, 3, 1 )

  ! (3,1,2) (no 2)
  allocate( dsp_p%next )
  dsp_p => dsp_p%next
  call new( dsp_p, isize )
  call permute( dsp_p, this, 3, 1, 2 )

  ! (3,2,1) (no 1)
  allocate( dsp_p%next )
  dsp_p => dsp_p%next
  call new( dsp_p, isize )
  call permute( dsp_p, this, 3, 1 )

  if ( present(last_p) ) then
    last_p => dsp_p
  endif

  return
end subroutine run_2


subroutine run_1( this, length, last_p )
!
! generate on-axis displacement vectors
!
! inout:
!   this : disp object
!
! input :
!   length : displacement (x-direction)
!
! output:
!   last_p : pointer to the last disp object in the list (optional)
!
  type(disp), intent(inout), target :: this
  integer, intent(in) :: length
  type(disp), intent(out), optional, pointer :: last_p
!------------------------------------------------------
  type(disp), pointer :: dsp_p
  integer :: j

  if ( length /= SIZE(this%dir) ) then
    call new( this, length )
  endif

  dsp_p => this

  ! x-direction
  call new( dsp_p, length )
  do j = 1, length
    dsp_p%dir(j) = 1
  enddo

  ! y-direction
  allocate( dsp_p%next )
  dsp_p => dsp_p%next
  call new( dsp_p, length )
  do j = 1, length
    dsp_p%dir(j) = 2
  enddo

  ! z-direction
  allocate( dsp_p%next )
  dsp_p => dsp_p%next
  call new( dsp_p, length )
  do j = 1, length
    dsp_p%dir(j) = 3
  enddo

  if ( present(last_p) ) then
    last_p => dsp_p
  endif

  return
end subroutine run_1

end module bresenham_module
