#include "configF90.h"
! generate a list of three-dimensional paths
!   (M.Hayakawa, November 28, 2010)

module path_generate_class
  use constants_module
!--- module "param_size" was defined in HMCLIB/chrolog_class.F
  use param_size
!  use error_class
  implicit none
  private
  
  public :: disp
  public :: disp_set
  public :: path_pattern
  public :: new
  public :: init ! change flag_all
  public :: delete
  public :: Bresenham
  public :: run
  public :: generate ! only for debug purpose

  ! 1 : displacement vectors with both negative and positive z-component 
  !     are generated
  ! 0 : Displacement vectors with only non-negative z-component are generated
  ! ("0" is needed when Wilson-loop is averaged over starting points)
  ! "flag_all" can be changed through the function "init"
  integer, save :: flag_all = 0

  ! default values of maximum and minimum values of R^2
  integer, parameter :: Rmax2_0 = 3 * ( MIN(NTX, NTY, NTZ) /2 )**2
  integer, parameter :: Rmin2_0 = 1

  type disp
    ! array of directions of displacement vectors {1, 2, 3, -1, -2, -3}
    !   (size of array = total number of steps)
    integer, allocatable :: dir(:) 
    ! pointer to next disp
    type(disp), pointer :: next => NULL()
  end type disp

  type disp_set
    ! step size in each direction, ic(1) >= ic(2) >= ic(3) >= 0
    integer :: ic(NDIM-1)
    ! chain of disp with the same (ic(1), ic(2), ic(3))
    type(disp) :: sv
    ! pointer to the next disp_set with the same R^2, where
    !   R^2 = ic(1)**2 + ic(2)**2 + ic(3)**2
    type(disp_set), pointer :: next => NULL()
  end type disp_set

  type path_pattern
    ! array of disp_set indexed by R^2, where
    !   R^2 = ic(1)**2 + ic(2)**2 + ic(3)**2
    type(disp_set), allocatable :: ss(:)
    ! maximum value of R^2
    integer :: Rmax2 = Rmax2_0
    ! minimum value of R^2
    integer :: Rmin2 = Rmin2_0
  end type path_pattern

  type disp_set_p
    type(disp_set), pointer :: p
  end type disp_set_p


  interface new
    module procedure new_disp
    module procedure new_path_pattern
  end interface

  interface delete
    module procedure delete_disp
    module procedure delete_disp_set
    module procedure delete_path_pattern
  end interface

  interface free
    module procedure free_disp
    module procedure free_disp_set
  end interface

  interface init
    module procedure init_disp
  end interface

  interface Bresenham
    module procedure Bresenham_2d
    module procedure Bresenham_3d
  end interface

  interface permute
    module procedure permute_2
    module procedure permute_3
  end interface

  interface sflip
    ! get new displacement vectors by flipping (specified) directions
    module procedure sflip_all
    module procedure sflip_1
    module procedure sflip_2
  end interface

  interface generate
    module procedure generate_all_signs
  end interface

  interface calc
    module procedure calc_R2
  end interface

  interface run
    module procedure run_3
    module procedure run_2
    module procedure run_1
  end interface

contains

subroutine new_disp( this, nstep )
!
! allocate v
!
! inout:
!   this : disp object
!
! inout:
!   nstep : number of steps = number of displacement vectors
!
  implicit none
  type(disp), intent(inout) :: this
  integer :: nstep

  if ( allocated(this%dir) ) deallocate(this%dir)
  allocate( this%dir(nstep) )

  return
end subroutine new_disp


subroutine new_path_pattern( this, Rmax2_in, Rmin2_in )
!
! initialization of path_pattern
!
! inout:
!   this : path_pattern object
!
! input (optional) :
!   Rmax2_in : maximum R^2
!   Rmin2_in : minimum R^2
!
  type(path_pattern), intent(inout) :: this
  integer, intent(in), optional :: Rmax2_in, Rmin2_in
!------------------------------------------

  if (present(Rmin2_in)) then
    if ( Rmin2_0 < Rmin2_in ) then 
      this%Rmin2 = Rmin2_in
    endif
  endif

  if (present(Rmax2_in)) then
    if ( Rmax2_in < Rmax2_0 ) then
      this%Rmax2 = Rmax2_in
    endif
  endif
  
  if (allocated(this%ss)) then
    deallocate(this%ss)
  endif

  allocate( this%ss( this%Rmin2:this%Rmax2 ) )

  return   
end subroutine new_path_pattern


subroutine delete_disp( this )
!
! deconstructor
!
! inout:
!   this : disp object
!
  implicit none
  type(disp), intent(inout) :: this
!--------------------------------------------------

  if ( allocated(this%dir) ) deallocate(this%dir)
  call free( this%next )

  return
end subroutine delete_disp


subroutine delete_disp_set( this )
!
! deconstructor
!
! inout:
!   this : disp_set object
!
  implicit none
  type(disp_set), intent(inout) :: this

  call delete( this%sv )
  call free( this%next )

  return
end subroutine delete_disp_set


subroutine delete_path_pattern( this )
!
! deconstructor
!
! inout:
!   this : path_pattern object
!
  implicit none
  type(path_pattern), intent(inout) :: this

  integer j

  if ( .not. allocated(this%ss) ) then
    return
  endif

  do j = this%Rmin2, this%Rmax2
    call delete( this%ss(j) )
  enddo

  deallocate( this%ss )

  return
end subroutine delete_path_pattern


recursive subroutine free_disp( sv )
!
! free v in all lists
!
! inout:
!   sv : disp object
!
  implicit none
  type(disp), intent(inout), pointer :: sv

  if ( associated(sv) ) then
    call free_disp( sv%next )
    deallocate( sv%dir )
    deallocate( sv )
!    write(*,*) "free !"  ! for debug purpose
  endif

  return
end subroutine free_disp


recursive subroutine free_disp_set( ss )
!
! free v in all lists
!
! inout:
!   ss : disp_set object
!
  implicit none
  type(disp_set), intent(inout), pointer :: ss

  if ( associated(ss) ) then
    call free_disp_set( ss%next )
    call delete( ss%sv )
    deallocate( ss )
!    write(*,*) "free !"  ! for debug purpose
  endif

  return
end subroutine free_disp_set


subroutine init_disp( this, flag_all_in )
!
! set flag_all
!
! input :
!   this        : disp obj (dummy)
!   flag_all_in : flag_all 
!                 (0 : restricted, 1 : allow steps in all negative directions)
!
  implicit none
  type(disp), intent(in) :: this
  integer, intent(in) :: flag_all_in

  flag_all = flag_all_in

  return
end subroutine init_disp


subroutine Bresenham_2d( this, lmax, lmin )
!
! generate displacement vectors via 2-dim verions of Bresenham algorithm 
!
! inout:
!   this : disp object
!
! input :
!   lmax : maximum size of displacement (x-direction)
!   lmin : minimum size of displacement (y-direction)
!
  type(disp), intent(inout) :: this
  integer, intent(in) :: lmax, lmin
  integer, parameter :: dir_max = 1, dir_min = 2
!----------------------------------------------------------
  integer :: lmax2, lmin2
  integer :: h
  integer :: iter
  integer :: istep, nstep

  nstep = lmax + lmin
  if ( nstep /= SIZE(this%dir) ) then
    call new( this, nstep )
  endif

  lmax2 = 2 *lmax
  lmin2 = 2 *lmin
  h = lmin2 - lmax

  istep = 0
  do iter = 1, lmax
    istep = istep + 1
    this%dir(istep) = dir_max ! max-direction
    if ( h >= 0 ) then
      h = h - lmax2
      istep = istep + 1
      this%dir(istep) = dir_min ! min-direction
    endif
    h = h + lmin2
  enddo ! iter
  
  return
end subroutine Bresenham_2d


subroutine Bresenham_3d( this, lmax, lmid, lmin )
!
! generate displacement vectors via 3-dim verions of Bresenham algorithm 
!
! inout:
!   this : disp object
!
! input :
!   lmax : maximum size of displacement (x-direction)
!   lmid : middle size of displacement  (y-direction)
!   lmin : minimum size of displacement (z-direction)
!
  type(disp), intent(inout) :: this
  integer, intent(in) :: lmax, lmid, lmin
  integer, parameter :: dir_max = 1, dir_mid = 2, dir_min = 3
!----------------------------------------------------------
  integer :: lmax2, lmid2, lmin2
  integer :: h_max_mid, h_max_min
  integer :: iter
  integer :: istep, nstep

  nstep = lmax + lmid + lmin
  if ( nstep /= SIZE(this%dir) ) then
    call new( this, nstep )
  endif

  lmax2 = 2 *lmax
  lmid2 = 2 *lmid
  lmin2 = 2 *lmin
  h_max_mid = lmid2 - lmax
  h_max_min = lmin2 - lmax

  istep = 0
  do iter = 1, lmax
    istep = istep + 1
    this%dir(istep) = dir_max ! max-direction
    if ( h_max_mid >= 0 ) then
      h_max_mid = h_max_mid - lmax2
      istep = istep + 1
      this%dir(istep) = dir_mid ! mid-direction
    endif
    if ( h_max_min >= 0 ) then
      h_max_min = h_max_min - lmax2
      istep = istep + 1
      this%dir(istep) = dir_min ! min-direction
    endif
    h_max_mid = h_max_mid + lmid2
    h_max_min = h_max_min + lmin2
  enddo ! iter
  
  return
end subroutine Bresenham_3d


subroutine permute_2( this, src, dir1, dir2 )
!
! generate new displacement vectors by dir1 <=> dir2 
!
! inout:
!   this : disp object
!
! input:
!   src  : disp object
!   dir1 : direction \in {1,2,3}
!   dir2 : direction (\ne dir1) \in {1,2,3}
!
  type(disp), intent(inout) :: this
  type(disp), intent(in) :: src
  integer, intent(in) :: dir1, dir2
!------------------------------------------------------
  integer :: j

  do j = 1, SIZE(this%dir)
    if ( src%dir(j) == dir1 ) then
      this%dir(j) = dir2
    else if ( src%dir(j) == dir2 ) then
      this%dir(j) = dir1
    else
      this%dir(j) = src%dir(j)
    endif
  enddo

  return
end subroutine permute_2


subroutine permute_3( this, src, dir1, dir2, dir3 )
!
! generate new displacement vectors by 
!   dir1 <= (1 in src), dir2 <= (2 in src) dir3 <= (3 in src)
!
! inout:
!   this : disp object
!
! input:
!   src  : disp object
!   dir1 : direction \in {1,2,3}
!   dir2 : (\ne dir1) \in {1,2,3}
!   dir3 : (\ne dir1, dir2) \in {1,2,3}
!
  type(disp), intent(inout) :: this
  type(disp), intent(in) :: src
  integer, intent(in) :: dir1, dir2, dir3
!------------------------------------------------------
  integer :: j

  do j = 1, SIZE(this%dir)
    if ( src%dir(j) == 1 ) then
      this%dir(j) = dir1
    else if ( src%dir(j) == 2 ) then
      this%dir(j) = dir2
    else
      this%dir(j) = dir3
    endif
  enddo

  return
end subroutine permute_3


subroutine sflip_all( this, src )
!
! generate new displacement vectors by flipping all directions 
!
! inout:
!   this : disp object
!
! input:
!   src  : disp object
!
  type(disp), intent(inout) :: this
  type(disp), intent(in) :: src
!------------------------------------------------------
  integer :: j

  do j = 1, SIZE(this%dir)
    this%dir(j) = - src%dir(j)
  enddo

  return
end subroutine sflip_all


subroutine sflip_1( this, src, dir )
!
! generate new displacement vectors by flipping the sign of direction, dir 
!
! inout:
!   this : disp object
!
! input:
!   src : disp object
!   dir : direction to be reversed
!
  type(disp), intent(inout) :: this
  type(disp), intent(in) :: src
  integer, intent(in) :: dir
!------------------------------------------------------
  integer :: j

  do j = 1, SIZE(this%dir)
    if ( src%dir(j) == dir ) then
      this%dir(j) = - src%dir(j)
    else
     this%dir(j) = src%dir(j)
    endif
  enddo

  return
end subroutine sflip_1


subroutine sflip_2( this, src, dir1, dir2 )
!
! generate new displacement vectors by flipping the signs of directions,
! dir1, dir2 
!
! inout:
!   this : disp object
!
! input:
!   src  : disp object
!   dir1 : direction to be reversed
!   dir2 : direction to be reversed (\ne dir1)
!
  type(disp), intent(inout) :: this
  type(disp), intent(in) :: src
  integer, intent(in) :: dir1, dir2
!------------------------------------------------------
  integer :: j

  do j = 1, SIZE(this%dir)
    if ( ( src%dir(j) == dir1 ) .or. ( src%dir(j) == dir2 ) ) then
      this%dir(j) = - src%dir(j)
    else
      this%dir(j) = src%dir(j)
    endif
  enddo

  return
end subroutine sflip_2


subroutine generate_all_signs( this, last_p )
!
! generate new displacement vectors by flipping the signs of directions
! in all possible ways
!
! (Caution: ) This should be used only when
! the original vectors contain all 3 directions
!
! inout:
!   this : disp object
!
! out :
!   last_p : (optional) pointer to the last disp object in the list
!
  type(disp), intent(inout), target :: this
  type(disp), pointer, optional :: last_p
!---------------------------------------------------------
  type(disp), pointer :: src, res
  integer :: isize

  src => this
  isize = SIZE(src%dir)

  ! find the last in the list
  res => src
  do while ( associated( res%next ) )
    res => res%next
  enddo

  ! 1 -> -1
  allocate( res%next )
  res => res%next
  call new( res, isize )
  call sflip( res, src, 1 )

  ! 2 -> -2
  allocate( res%next )
  res => res%next
  call new( res, isize )
  call sflip( res, src, 2 )

  ! 3 -> -3
  if ( flag_all /= 0 ) then
    allocate( res%next )
    res => res%next
    call new( res, isize )
    call sflip( res, src, 3 )
  endif

  ! 1 -> -1, 2 -> -2
  allocate( res%next )
  res => res%next
  call new( res, isize )
  call sflip( res, src, 1, 2 )

  ! 1 -> -1, 3 -> -3
  if ( flag_all /= 0 ) then
    allocate( res%next )
    res => res%next
    call new( res, isize )
    call sflip( res, src, 1, 3 )
  endif

  ! 2 -> -2, 3 -> -3
  if ( flag_all /= 0 ) then
    allocate( res%next )
    res => res%next
    call new( res, isize )
    call sflip( res, src, 2, 3 )
  endif

  ! 1 -> -1, 2 -> -2, 3 -> -3
  if ( flag_all /= 0 ) then
    allocate( res%next )
    res => res%next
    call new( res, isize )
    call sflip( res, src )
  endif

  if (present(last_p)) then
    last_p => res
  endif

  return
end subroutine generate_all_signs


subroutine run_3( this, lmax, lmid, lmin )
!
! generate all possible displacement vectors 
! upon the displacement vectors obtained via 3-dim verions
! of Bresenham algorithm
!
! Caution : 
!  (1) only use for lmin \ne 0
!      (Otherwise, generate patterns includes the same ones)
!  (2) user is responsible to correctly give lmax >= lmid >= lmin
!
! inout:
!   this : disp object
!
! input :
!   lmax : maximum size of displacement (x-direction)
!   lmid : middle size of displacement  (y-direction)
!   lmin : minimum size of displacement (z-direction)
!
  type(disp), intent(inout) :: this
  integer, intent(in) :: lmax, lmid, lmin
!------------------------------------------------------
  type(disp), pointer :: sv_p, res
  integer :: isize

  ! generate basic steps 
  call Bresenham( this, lmax, lmid, lmin )

  ! generate step vectors from basic steps by reversing directions
  ! in all possible ways
  call generate( this, sv_p )

  isize = SIZE(this%dir)

  ! (1,3,2)
  allocate( sv_p%next )
  res => sv_p%next
  call new( res, isize )
  call permute( res, this, 2, 3 )
  call generate( res, sv_p )

  ! (2,1,3)
  allocate( sv_p%next )
  res => sv_p%next
  call new( res, isize )
  call permute( res, this, 1, 2 )
  call generate( res, sv_p )

  ! (2,3,1) 
  allocate( sv_p%next )
  res => sv_p%next
  call new( res, isize )
  call permute( res, this, 2, 3, 1 )
  call generate( res, sv_p )

  ! (3,1,2)
  allocate( sv_p%next )
  res => sv_p%next
  call new( res, isize )
  call permute( res, this, 3, 1, 2 )
  call generate( res, sv_p )

  ! (3,2,1)
  allocate( sv_p%next )
  res => sv_p%next
  call new( res, isize )
  call permute( res, this, 3, 1 )
  call generate( res, sv_p )

  return
end subroutine run_3


subroutine run_2( this, lmax, lmin )
!
! generate all possible displacement vectors 
! upon the displacement vectors obtained via 3-dim verions
! of Bresenham algorithm
!
! Caution : 
!  (1) only use for lmin \ne 0
!      (Otherwise, generate patterns includes the same ones)
!  (2) user is responsible to correctly give lmax >= lmin
!
! inout:
!   this : disp object
!
! input :
!   lmax : maximum size of displacement (x-direction)
!   lmin : middle size of displacement  (y-direction)
!
  type(disp), intent(inout), target :: this
  integer, intent(in) :: lmax, lmin
!------------------------------------------------------
  type(disp), pointer :: sv_p, res
  integer :: isize

  isize = SIZE(this%dir)

  ! generate basic steps 
  call Bresenham( this, lmax, lmin )

  sv_p => this
  ! generate step vectors from basic steps by reversing directions
  ! in all possible ways
  ! 1 -> -1
  allocate( sv_p%next )
  res => sv_p%next
  call new( res, isize )
  call sflip( res, sv_p, 1 )
  if ( flag_all /= 0 ) then
    ! 2 -> -2
    allocate( res%next )
    res => res%next
    call new( res, isize )
    call sflip( res, sv_p, 2 )
    ! all
    allocate( res%next )
    res => res%next
    call new( res, isize )
    call sflip( res, sv_p )
  endif

  ! (1,3,2) (no 2)
  allocate( res%next )
  sv_p => res%next
  call new( sv_p, isize )
  call permute( sv_p, this, 2, 3 )
  ! 1 -> -1
  allocate( sv_p%next )
  res => sv_p%next
  call new( res, isize )
  call sflip( res, sv_p, 1 )
  if ( flag_all /= 0 ) then
    ! 3 -> -3
    allocate( res%next )
    res => res%next
    call new( res, isize )
    call sflip( res, sv_p, 3 )
    ! all
    allocate( res%next )
    res => res%next
    call new( res, isize )
    call sflip( res, sv_p )
  endif

  ! (2,1,3) (no 3)
  allocate( res%next )
  sv_p => res%next
  call new( sv_p, isize )
  call permute( sv_p, this, 1, 2 )
  ! 1 -> -1
  allocate( sv_p%next )
  res => sv_p%next
  call new( res, isize )
  call sflip( res, sv_p, 1 )
  if ( flag_all /= 0 ) then
    ! 2 -> -2
    allocate( res%next )
    res => res%next
    call new( res, isize )
    call sflip( res, sv_p, 2 )
    ! all
    allocate( res%next )
    res => res%next
    call new( res, isize )
    call sflip( res, sv_p )
  endif

  ! (2,3,1) (no 1)
  allocate( res%next )
  sv_p => res%next
  call new( sv_p, isize )
  call permute( sv_p, this, 2, 3, 1 )
  if ( flag_all /= 0 ) then
    ! 2 -> -2
    allocate( sv_p%next )
    res => sv_p%next
    call new( res, isize )
    call sflip( res, sv_p, 2 )
    ! 3 -> -3
    allocate( res%next )
    res => res%next
    call new( res, isize )
    call sflip( res, sv_p, 3 )
    ! all
    allocate( res%next )
    res => res%next
    call new( res, isize )
    call sflip( res, sv_p )
  endif

  ! (3,1,2) (no 2)
  allocate( res%next )
  sv_p => res%next
  call new( sv_p, isize )
  call permute( sv_p, this, 3, 1, 2 )
  ! 1 -> -1
  allocate( sv_p%next )
  res => sv_p%next
  call new( res, isize )
  call sflip( res, sv_p, 1 )
  if ( flag_all /= 0 ) then
    ! 3 -> -3
    allocate( res%next )
    res => res%next
    call new( res, isize )
    call sflip( res, sv_p, 3 )
    ! all
    allocate( res%next )
    res => res%next
    call new( res, isize )
    call sflip( res, sv_p )
  endif

  ! (3,2,1) (no 1)
  allocate( res%next )
  sv_p => res%next
  call new( sv_p, isize )
  call permute( sv_p, this, 3, 1 )
  if ( flag_all /= 0 ) then
    ! 2 -> -2
    allocate( sv_p%next )
    res => sv_p%next
    call new( res, isize )
    call sflip( res, sv_p, 2 )
    ! 3 -> -3
    allocate( res%next )
    res => res%next
    call new( res, isize )
    call sflip( res, sv_p, 3 )
    ! all
    allocate( res%next )
    res => res%next
    call new( res, isize )
    call sflip( res, sv_p )
  endif

  return
end subroutine run_2


subroutine run_1( this, length )
!
! generate on-axis displacement vectors
!
! inout:
!   this : disp object
!
! input :
!   length : displacement (x-direction)
!
  type(disp), intent(inout), target :: this
  integer, intent(in) :: length
!------------------------------------------------------
  type(disp), pointer :: sv_p
  integer :: j

  if ( length /= SIZE(this%dir) ) then
    call new( this, length )
  endif

  sv_p => this

  ! x-direction
  do j = 1, length
    sv_p%dir(j) = 1
  enddo

  ! sign flips
  if ( flag_all /= 0 ) then
    allocate( sv_p%next )
    sv_p => sv_p%next
    do j = 1, length
      sv_p%dir(j) = -1
    enddo
  endif

  ! y-direction
  allocate( sv_p%next )
  sv_p => sv_p%next
  do j = 1, length
    sv_p%dir(j) = 2
  enddo

  ! sign flips
  if ( flag_all /= 0 ) then
    allocate( sv_p%next )
    sv_p => sv_p%next
    do j = 1, length
      sv_p%dir(j) = -2
    enddo
  endif

  ! z-direction
  allocate( sv_p%next )
  sv_p => sv_p%next
  do j = 1, length
    sv_p%dir(j) = 3
  enddo

  ! sign flips
  if ( flag_all /= 0 ) then
    allocate( sv_p%next )
    sv_p => sv_p%next
    do j = 1, length
      sv_p%dir(j) = -3
    enddo
  endif

  return
end subroutine run_1


subroutine calc_R2( this )
!
! Generate chains of all possible (Cmax, Cmid. Cmin) for each R2
!
! inout:
!   this : path_pattern object
!
  implicit none
  type(path_pattern), intent(inout), target :: this
!-------------------------------------------------
  type(disp_set_p), allocatable :: pt(:)
  real(DP) :: xr
  integer, allocatable :: iflag(:)
  integer :: j1, j2, j3
  integer :: n1l, n2l, n3l
  integer :: n1u
  integer :: k1_2, k12_2
  integer :: r2, Rmin2, Rmax2

  Rmin2 = this%Rmin2
  Rmax2 = this%Rmax2
  allocate(    pt( Rmin2:Rmax2 ) )
  allocate( iflag( Rmin2:Rmax2 ) )

  do r2 = Rmin2, Rmax2
    pt(r2)%p => this%ss(r2)
    iflag(r2) = 0
  enddo

  n1l = 1
  n2l = 0
  n3l = 0

  r2 = AINT(SQRT(REAL(Rmax2,kind=DP)),kind=DP)
  n1u = MIN( r2 + 1, NTX /2, NTY /2, NTZ /2 )

  do j1 = n1l, n1u
    k1_2 = j1**2
    do j2 = n2l, j1
      k12_2 = k1_2 + j2**2
      do j3 = n3l, j2
        r2 = k12_2 + j3**2
        if ( ( r2 < Rmin2 ) .or. ( Rmax2 < r2 ) ) cycle
        pt(r2)%p%ic(1) = j1
        pt(r2)%p%ic(2) = j2
        pt(r2)%p%ic(3) = j3
        ! for the next
        allocate( pt(r2)%p%next )
        pt(r2)%p => pt(r2)%p%next
        iflag(r2) = 1
      enddo ! j3
    enddo ! j2
  enddo ! j1

  do r2 = Rmin2, Rmax2
    ! discard unused variable
    if ( iflag(r2) == 1 ) then
      deallocate( pt(r2)%p )
    endif
  enddo

  deallocate( iflag, pt )

  return
end subroutine calc_R2


end module path_generate_class
